
from pbipinspect.pbip import Pbip
from pbipinspect.utils import (
    check_lines_max_length,
    smart_join
)
from textwrap import dedent
from typing import Callable, Literal

ExpectState = Literal['Warning', 'Error', 'Info']
ExpectMessage = dict[str, str]
ColTypes = Literal[
    'double',
    'percentage', # not supported
    'duration',   # not supported
    'decimal',
    'int64',
    'int32',
    'int16',
    'int8',

    'string',

    'dateTime',
    'time'       # not supported

    'boolean',

    'binary',    # not supported
    'list',      # not supported
    'record'     # not supported
]


def expect_col_starts_with(
    pattern: str, 
    col_type: list[ColTypes] | ColTypes | None,
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if columns in a Pbip object start
    with the specified pattern and have a supported data type.

    Parameters
    ----------
    pattern : str
        The pattern to match against the column name.
    col_type : list[ColTypes] | ColTypes | None
        The data type(s) to check for.
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if columns in a Pbip object start
        with the specified pattern and have a supported data type.
    """
    not_supported = ['binary', 'list', 'record', 'time', 'percentage', 'duration'] 
    
    if col_type and not isinstance(col_type, list):
        col_type = [col_type]

    if col_type:
        for col in col_type:
            if col in not_supported:
                Warning(f"Col type '{col}' is not supported. I'll ignore this type.")
        col_type = [x for x in col_type if x not in not_supported]

    def function(pbip: Pbip) -> list[ExpectMessage]:
        """
        Checks if columns in each table of the Pbip object start with a specified pattern
        and match any of the specified data types.

        Args:
            pbip (Pbip): The Pbip object containing tables and columns to check.

        Returns:
            list[ExpectMessage]: A list of messages detailing any columns that do not meet the criteria.
        """
        tables = pbip.tables
        checks = []
        for table in tables:
            table_name = table['name']
            cols = pbip.get_table_field(table_name, 'columns')
            if not cols:
                continue
            for col in cols:
                col_name = col['name']
                check_type = col_type is None or col['dataType'] in col_type
                if check_type and not col_name.startswith(pattern):
                    checks.append({
                        'expect': 'expect_col_starts_with',
                        'state': state,
                        'message': f"Column '{col_name}' in table '{table_name}' must start with '{pattern}'"
                    })
        return checks

    return function

def expect_measure_starts_with(
    pattern: str, 
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if measures in a Pbip object start
    with the specified pattern.

    Parameters
    ----------
    pattern : str
        The pattern to match against the measure name.
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if measures in a Pbip object start
        with the specified pattern.
    """
    
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        for table in tables:
            table_name = table['name']
            measures = pbip.get_table_field(table_name, 'measures')
            if not measures:
                continue
            for measure in measures:
                measure_name = measure['name']
                if not measure_name.startswith(pattern):
                    checks.append({
                        'expect': 'expect_measure_starts_with',
                        'state': state,
                        'message': f"Measure '{measure_name}' in table '{table_name}' must start with '{pattern}'",
                    })
        return checks
    return function

def expect_table_starts_with(
    pattern: str, 
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if table names in a Pbip object start
    with the specified pattern.

    Parameters
    ----------
    pattern : str
        The pattern to match against the table name.
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if table names in a Pbip object start
        with the specified pattern.
    """
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        for table in tables:
            table_name = table['name']
            if not table_name.startswith(pattern):
                checks.append({
                    'expect': 'expect_table_starts_with',
                    'state': state,
                    'message': f"Table '{table_name}' must start with '{pattern}'",
                })
        return checks
    return function

def expect_cols_in_relationship_has_same_type(
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if columns in relationships of a Pbip object have the same type.

    Parameters
    ----------
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if columns in relationships of a Pbip object have the same type.
    """
    def function(pbip: Pbip) -> list[ExpectMessage]:
        checks = []
        relationships = pbip.relationships
        for relationship in relationships:
            from_column = relationship['fromColumn']
            from_table = relationship['fromTable']
            to_column = relationship['toColumn']
            to_table = relationship['toTable']

            from_table_column = pbip.get_table_column(from_table, from_column)
            to_table_column = pbip.get_table_column(to_table, to_column)

            if from_table_column and to_table_column:
                from_column_type = from_table_column['dataType']
                to_column_type = to_table_column['dataType']

                if from_column_type != to_column_type:
                    checks.append({
                        'expect': 'expect_cols_in_relationship_has_same_type',
                        'state': state,
                        'message': dedent(f"""
                            Column '{from_column}' in table '{from_table}' and '{to_column}' in table '{to_table}' must have the same type.
                            But I found '{from_column}' with '{from_column_type}' type and '{to_column}' with '{to_column_type}' type.
                        """).strip()
                    })
        return checks
    return function

def expect_table_name_no_spaces(
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if table names in a Pbip object do not contain spaces.

    Parameters
    ----------
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if table names in a Pbip object do not contain spaces.
    """
    
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        for table in tables:
            table_name = table['name']
            if ' ' in table_name:
                checks.append({
                    'expect': 'expect_table_name_no_spaces',
                    'state': state,
                    'message': f"Table '{table_name}' must not contain spaces",
                })
        return checks
    return function

def expect_dax_lines_length(
    max_length: int,
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if DAX measures in a Pbip object do not exceed the specified maximum length.

    Parameters
    ----------
    max_length : int
        The maximum length of a DAX measure.
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if DAX measures in a Pbip object do not exceed the specified maximum length.
    """
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        for table in tables:
            table_name = table['name']
            measures = pbip.get_table_field(table_name, 'measures')
            if not measures:
                continue
            for measure in measures:
                measure_name = measure['name']
                expression = measure['expression']
                lines = check_lines_max_length(expression, max_length)
                text = smart_join([x[0] for x in lines])
                if any([x[1] for x in lines]):
                    checks.append({
                        'expect': 'expect_dax_lines_length',
                        'state': state,
                        'message': f"Measure '{measure_name}' in table '{table_name}' has line(s) {text} longer than {max_length} characters",
                    })
        return checks
    return function

def expect_m_lines_length(
    max_length: int,
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if M language expressions in the partitions
    of tables within a Pbip object do not exceed the specified maximum length.

    Parameters
    ----------
    max_length : int
        The maximum length of a line in an M language expression.
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if M language expressions in the partitions
        of tables within a Pbip object do not exceed the specified maximum length.
    """
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        for table in tables:
            table_name = table['name']
            partitions = pbip.get_table_field(table_name, 'partitions')
            if not partitions:
                continue
            for partition in partitions:
                expression = partition['expression']
                lines = check_lines_max_length(expression, max_length)
                text = text = smart_join([x[0] for x in lines])
                if any([x[1] for x in lines]):
                    checks.append({
                        'expect': 'expect_m_lines_length',
                        'state': state,
                        'message': f"Source code of table '{table_name}' has line(s) {text} longer than {max_length} characters",
                    })
        return checks
    return function

def expect_measures_in_specific_table(
    table_name: str,
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if measures in a Pbip object are in a specific table.

    Parameters
    ----------
    table_name : str
        The name of the table where measures should be.
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if measures in a Pbip object are in a specific table.
    """
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        tables_with_measures = []
        for table in tables:
            measures = pbip.get_table_field(table['name'], 'measures')
            if measures:
                tables_with_measures.append(table['name'])
        invalid_tables = [
            table
            for table in tables_with_measures
            if table != table_name
        ]
        if invalid_tables:
            checks.append({
                'expect': 'expect_measures_in_specific_table',
                'state': state,
                'message': f"Measures must be in table '{table_name}' but found in table(s) {smart_join(invalid_tables)}",
            })
        return checks
    return function

def expect_no_calculated_columns(
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if calculated columns are not present in any table of a Pbip object.

    Parameters
    ----------
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if calculated columns are not present in any table of a Pbip object.
    """
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        for table in tables:
            columns = []
            table_name = table['name']
            table_columns = pbip.get_table_field(table_name, 'columns')
            if table_columns is None:
                continue
            for column in table_columns:
                if column['calculated']:
                    columns.append(column['name'])
            if columns:
                checks.append({
                    'expect': 'expect_no_calculated_columns',
                    'state': state,
                    'message': f"Table '{table_name}' has calculated columns: {smart_join(columns)}",
                })
        return checks
    return function

def expect_all_relationships_active(
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if all relationships in a Pbip object are active.

    Parameters
    ----------
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if all relationships in a Pbip object are active.
    """

    def function(pbip: Pbip) -> list[ExpectMessage]:
        relationships = pbip.relationships
        if not relationships:
            return []
        checks = []
        for relationship in relationships:
            from_table = relationship['fromTable']
            to_table = relationship['toTable']
            from_column = relationship['fromColumn']
            to_column = relationship['toColumn']
            if not relationship['isActive']:
                checks.append({
                    'expect': 'expect_all_relationships_active',
                    'state': state,
                    'message': (f"Relationship between '{from_table}.{from_column}' and '{to_table}.{to_column}' must be active."),
                })
        return checks
    return function

def expect_no_calculated_tables(
    state: ExpectState = 'Warning'
) -> Callable:
    """
    Returns a function that checks if calculated tables are not present in a Pbip object.

    Parameters
    ----------
    state : ExpectState, optional
        The state of the expectation, defaults to 'Warning'.

    Returns
    -------
    Callable
        A function that checks if calculated tables are not present in a Pbip object.
    """
    def function(pbip: Pbip) -> list[ExpectMessage]:
        tables = pbip.tables
        checks = []
        for table in tables:
            if table['partitions'][0]['type'] == 'calculated':
                checks.append({
                    'expect': 'expect_no_calculated_tables',
                    'state': state,
                    'message': f"Table '{table['name']}' is calculated",
                })
        return checks
    return function




DEFAULT_TEMPLATE = """
{% if metadata -%}- [Overview](#Overview){% endif %}
- [Expectations](#Expectations)
- [Model](#Model)
    - [Relationships](#Relationships)
    - [Expressions](#Expressions)
    - [Tables](#Tables)
        | Table name | Source code | Columns | Measures |
        | ---------- | ----------- | ------- | -------- |
    {%- for table in tables %}
        | [{{ table.name }}](#{{ sanitize_id(table.name) }}) |[Source code](#source-code-{{ sanitize_id(table.name) }}) | [Columns](#columns-{{ sanitize_id(table.name) }}) | [Measures](#measures-{{ sanitize_id(table.name) }}) |
    {%- endfor %}

----
{% if metadata %}
# <span id="Overview">Overview</span>
{% for md in metadata %}
## {{ md }}
{{ metadata[md] }}
{% endfor %}
----
{% endif -%}

# <span id="Expectations">Expectations</span>
{% if expects %}
| State | Expect | Message |
| ----- | ------ | ------- |
{%- for expect in expects %}
| {{ expect.state }} | {{ expect.expect }} | {{ expect.message }} |
{%- endfor %}
{% endif %}
{% if not expects %}Without warnings or errors{% endif %}

----

# <span id="Model">Model</span>
## <span id="Relationships">Relationships</span>
```mermaid
---
config:
    layout: elk
---
{{ mermaid_relationships }}
```

## <span id="Tables">Tables</span>
{% for table in tables %}
### <span id="{{ sanitize_id(table.name) }}">{{ table.name }} {% if table.isHidden %}(Hidden){% endif %} {% if table.isPrivate %}(Private){% endif %}</span>
{{ table.partitions[0].description }}
#### <span id="source-code-{{ sanitize_id(table.name) }}">Source code</span>
```m
{{ table.partitions[0].expression }}
```
#### <span id="columns-{{ sanitize_id(table.name) }}">Columns</span>
{%- if table.columns %}
| Column | Description | Type | Calculated | Is hidden |
| ------ | ----------- | ---- | ---------- | --------- |
{%- for column in table.columns %}
| {{ column.name }} | {{ column.description }} | {{ column.dataType }} | {{ column.calculated }} | {{ column.isHidden }} |
{%- endfor %}
{% endif %}
{% if not table.columns %}No columns{% endif %}
#### <span id="measures-{{ sanitize_id(table.name) }}">Measures</span>
{% if not table.measures %}No measures{% endif %}
{%- if table.measures %}
{%- for measure in table.measures %}
##### <span id="{{ sanitize_id(table.name) }}-{{ sanitize_id(measure.name) }}">{{ measure.name }}</span>
Description: {{ measure.description }}

Display folder: {{ measure.displayFolder }}

Dependencies:
{% for reference in measure.references -%}
{%- if reference.type == 'table' -%}
- {{ reference.name }}[{{ reference.column }}]
{% endif %}
{%- if reference.type == 'measure' -%}
- [{{ reference.name }}]
{%- endif %}
{% endfor %}

Expression:
```dax
{{ measure.expression }}
```
{% endfor %}
{% endif %}
{% endfor %}

## <span id="Expressions">Expressions</span>
{% if not expressions %}No expressions{% endif %}
{%- if expressions %}
{%- for expression in expressions %}
### <span id="{{ sanitize_id(expression.name) }}">{{ expression.name }}</span>
{{ expression.description }}

Type: {{ expression.type }}
{% if expression.type == 'function' %}
```m
{{ expression.expression }}
```
{% endif %}
{% if expression.type == 'expression' %}
Value: {{ expression.expression }}
{% endif %}
{% endfor %}
{% endif %}
"""




from copy import deepcopy
import re

def remove_doc_comments(x: str) -> str:
    """Removes Power BI style doc comments from a string.

    The function removes all occurrences of /* @doc */ comments from a given string.

    Parameters
    ----------
    x : str
        The string to remove the doc comments from.

    Returns
    -------
    str
        The string with the doc comments removed.
    """
    pattern = r'/\*\s?@doc.*?\*/'
    return re.sub(pattern, '', x, flags=re.DOTALL).strip()

def get_measure_references_table_column(expression: str) -> list[str]:
    """
    Returns all table column references in a given measure expression.

    The function returns all table column references found in a given measure expression.
    The table column references are returned as a list of strings, where each string
    represents a table column reference in the following format: table_name[column_name].

    Parameters
    ----------
    expression : str
        The measure expression to find the table column references in.

    Returns
    -------
    list[str]
        The list of table column references found in the measure expression.
    """
    pattern = r"(?:\b[a-zA-Z_][a-zA-Z0-9_]*\[[^\]]+\])|(?:\'[^\']+\'\[[^\]]+\])"
    matches = re.findall(pattern, expression)
    unique_matches = list(set(matches))
    return unique_matches

def get_measure_references_measure(expression: str) -> list[str]:
    """
    Returns all measure references in a given measure expression.

    The function returns all measure references found in a given measure expression.
    The measure references are returned as a list of strings, where each string
    represents a measure reference in the following format: measure_name.

    Parameters
    ----------
    expression : str
        The measure expression to find the measure references in.

    Returns
    -------
    list[str]
        The list of measure references found in the measure expression.
    """
    pattern = r'(?<![\w\'])\[[^\]]+\](?!\[\w)'
    matches = re.findall(pattern, expression)
    unique_matches = list(set([m.replace('[', '').replace(']', '') for m in matches]))
    return unique_matches

def get_measure_references(expression: str) -> list[dict]:
    """
    Extracts table column and measure references from a given expression.

    This function analyzes a measure expression to identify references to
    table columns and other measures. It returns a list of dictionaries
    where each dictionary describes a reference, specifying whether it is
    a 'table' or a 'measure', along with the respective name and column
    information for table references.

    Parameters
    ----------
    expression : str
        The measure expression to analyze for references.

    Returns
    -------
    list[dict]
        A list of dictionaries representing the references found in the
        expression. Each dictionary contains:
        - 'type': 'table' or 'measure'
        - 'name': Name of the table or measure
        - 'column': Name of the column (only for table references)
    """
    table_column = get_measure_references_table_column(expression)
    measures = get_measure_references_measure(expression)

    references = []
    for tc in table_column:
        table, column, _ = re.split('\\[|\\]', tc)
        references.append({
            'type': 'table',
            'name': table,
            'column': column
        })

    for m in measures:
        references.append({
            'type': 'measure',
            'name': m
        })
    return references

def get_table_description(x: str, pattern = r'@doc\s+(.*?)(?=@col|$)') -> str:
    """
    Extracts the table description from a given string.

    This function takes a string and returns the table description, if present.
    The table description is the text that follows the '@doc' tag until the
    '@col' tag or the end of the string.

    Parameters
    ----------
    x : str
        The string to extract the table description from.
    pattern : str, optional
        The regular expression pattern to match the table description.
    
    Returns
    -------
    str
        The extracted table description, or an empty string if no description is present.
    """
    description = re.search(pattern, x, flags=re.DOTALL)
    if description is None:
        return ''
    return description.group(1).replace('\t', '').strip()

def get_measure_description(x: str) -> str:
    """
    Extracts the measure description from a given string.

    This function takes a string and returns the measure description, if present.

    Parameters
    ----------
    x : str
        The string to extract the measure description from.

    Returns
    -------
    str
        The extracted measure description, or an empty string if no description is present.
    """
    pattern = r'/\* @doc(.*?)\*/'
    description = re.search(pattern, x, flags=re.DOTALL)
    if description is None:
        return ''
    return description.group(1).replace('\t', '').strip()

def get_table_column_description(x: str) -> dict[str, str]:
    """
    Extracts the column descriptions from a given string.

    This function takes a string and returns a dictionary with the column names as
    keys and their descriptions as values. The descriptions are the text that follows
    the '@col' tag until the '@col' or '*/' tag or the end of the string.

    Parameters
    ----------
    x : str
        The string to extract the column descriptions from.

    Returns
    -------
    dict[str, str]
        A dictionary with the column names as keys and their descriptions as values.
    """
    pattern = r'@col\s+(\w+):\s+(.*?)(?=@col|\*\/)'
    matches = re.findall(pattern, x, re.DOTALL)
    result = {match[0]: match[1].strip() for match in matches}
    return result

def put_column_description(
    columns: list[dict],
    table_expression: str
) -> list[dict]:
    """
    Adds column descriptions to a list of columns.

    This function takes a list of columns and a table expression, extracts the column
    descriptions from the table expression, and adds them to the columns.

    Parameters
    ----------
    columns : list[dict]
        The list of columns to add descriptions to.
    table_expression : str
        The table expression to extract the column descriptions from.

    Returns
    -------
    list[dict]
        The list of columns with descriptions added.
    """
    cur_columns = deepcopy(columns)
    columns_description = get_table_column_description(table_expression)
    if columns:
        for column in cur_columns:
            column_name = column['name']
            if column_name in columns_description:
                column['description'] = columns_description[column_name]
            else:
                column['description'] = ''
    return cur_columns

def get_relationship_filter_cardinality(
    filtering: str,
    to_cardinality: str | None,
    from_cardinality: str | None,
) -> tuple[str, str, str]:
    """
    Returns a tuple of 3 strings that represent the filtering direction, and the
    cardinality of the "from" and "to" columns in a relationship.

    Parameters
    ----------
    filtering : str
        The filtering behavior of the relationship. Can be 'singleDirection' or
        'bothDirections'.
    to_cardinality : str | None
        The cardinality of the "to" column in the relationship. Can be 'one' or
        'many', or None.
    from_cardinality : str | None
        The cardinality of the "from" column in the relationship. Can be 'one' or
        'many', or None.

    Returns
    -------
    tuple[str, str, str]
        A tuple of 3 strings. The first element is the filtering direction
        symbol, the second element is the "from" column cardinality symbol, and
        the third element is the "to" column cardinality symbol. The filtering
        direction symbol is '<' for single direction and '<>' for both
        directions. The cardinality symbols are '1' for one-to-one and '*' for
        one-to-many or many-to-many.
    """
    filtering_symbol = '<' if filtering == 'singleDirection' else '<>'
    from_cardinality_symbol = '*'
    to_cardinality_symbol = '*'

    if filtering == 'bothDirections':
        if from_cardinality == 'one' or to_cardinality == 'one':
            from_cardinality_symbol = '1'
            to_cardinality_symbol = '1'
        elif not from_cardinality and not to_cardinality:
            to_cardinality_symbol = '1'
        elif to_cardinality == 'oneToMany':
            to_cardinality_symbol = '1'

    elif filtering == 'singleDirection':
        if not to_cardinality:
            to_cardinality_symbol = '1'
        elif to_cardinality == 'many':
            from_cardinality_symbol = '*'
            to_cardinality_symbol = '*'
        elif to_cardinality == 'oneToMany':
            from_cardinality_symbol = '*'
            to_cardinality_symbol = '1'
    
    return (filtering_symbol, from_cardinality_symbol, to_cardinality_symbol)

def get_content_until_next_level(
    content: list[str],
    level_identifier: str='\t'
) -> list[str]:
    """
    Returns the content until the next level of indentation.

    Parameters
    ----------
    content : list[str]
        The content to be processed.
    level_identifier : str, optional
        The string used to identify the level of indentation. Defaults to '\t'.

    Returns
    -------
    list[str]
        The content until the next level of indentation.
    """
    if not content:
        return []
    last = content[-1].split('\n')
    idx_next_level = [
        idx for idx, x in enumerate(last)
        if x.count(level_identifier) == 1
    ]
    if not idx_next_level:
        return content[:-1] + ['\n'.join(last)]
    return content[:-1] + ['\n'.join(last[0:idx_next_level[0]])]

def get_table_property_content(
    tmdl: str, 
    property: str, 
    level_identifier: str='\t'
) -> list[str]:
    """
    Returns the content of a given table property from a TMDL file.

    Parameters
    ----------
    tmdl : str
        The content of the TMDL file.
    property : str
        The name of the property to be extracted. Can be 'column', 'measure', or
        'partition'.
    level_identifier : str, optional
        The string used to identify the level of indentation. Defaults to '\t'.

    Returns
    -------
    list[str]
        The content of the given table property.
    """
    content = tmdl.split(f'{level_identifier}{property} ')[1:]
    valid_content = get_content_until_next_level(content)
    return valid_content




from pathlib import Path
from pbipinspect.utils import get_regex_group
from pbipinspect.parse.utils import (
    get_measure_references,
    put_column_description,
    get_measure_description,
    get_relationship_filter_cardinality,
    get_table_description,
    get_table_property_content,
    remove_doc_comments,
)
import re
from typing import Optional

def build_component_path(semantic_model_path: str | Path, component: str) -> Path:
    new_path = Path(semantic_model_path)
    table_folder = Path.joinpath(new_path, 'definition', component)
    return table_folder

def build_relationship_path(semantic_model_path: str | Path) -> Path:
    return build_component_path(semantic_model_path, 'relationships.tmdl')

def build_expression_path(semantic_model_path: str | Path) -> Path:
    return build_component_path(semantic_model_path, 'expressions.tmdl')

def has_relationship_file(semantic_model_path: str | Path) -> bool:
    new_path = Path(semantic_model_path)
    relationship_file = build_relationship_path(new_path)
    if relationship_file.exists():
        return True
    return False

def has_expression_file(semantic_model_path: str | Path) -> bool:
    new_path = Path(semantic_model_path)
    expression_file = build_expression_path(new_path)
    if expression_file.exists():
        return True
    return False

def get_tmdl_relationship_file(semantic_model_path: str | Path) -> Path:
    new_path = Path(semantic_model_path)
    relationship_file = build_relationship_path(new_path)
    return relationship_file

def get_tmdl_expression_file(semantic_model_path: str | Path) -> Path:
    new_path = Path(semantic_model_path)
    expression_file = build_expression_path(new_path)
    return expression_file

def get_lineage_tag(x: str) -> Optional[str]:
    pattern = r'lineageTag:\s?([0-9a-zA-Z-]+)'
    tag = get_regex_group(x=x, pattern=pattern)
    return tag

def get_relationship_name(x: str) -> str:
    name = x.split('\n')[0]
    return name

def get_relationship_from_to(
    x: str, 
    pattern: str | re.Pattern[str]
) -> tuple[Optional[str], Optional[str]]:
    from_column = re.search(pattern, x)
    if from_column is None:
        return (None, None)
    else: 
        content = from_column.group(1).split('.')
        table = content[0]
        column = content[1]
        return (table, column)

def get_relationship_from_column(x: str) -> tuple[Optional[str], Optional[str]]:
    return get_relationship_from_to(x, r'fromColumn:\s?(.*)')
    
def get_relationship_to_column(x: str) -> tuple[Optional[str], Optional[str]]:
    return get_relationship_from_to(x, r'toColumn:\s?(.*)')

def get_relationship_to_cardinality(x: str) -> str:
    pattern = r'toCardinality:\s?(\w+)'
    cardinality = get_regex_group(x=x, pattern=pattern, if_none='oneToMany')
    return cardinality

def get_relationship_from_cardinality(x: str) -> str:
    pattern = r'fromCardinality:\s?(\w+)'
    cardinality = get_regex_group(x=x, pattern=pattern, if_none='oneToMany')
    return cardinality

def get_relationship_cross_filtering_behavior(x: str) -> str:
    pattern = r'crossFilteringBehavior:\s?(\w+)'
    behavior = get_regex_group(x=x, pattern=pattern, if_none='singleDirection')
    return behavior

def get_relationship_is_active(x: str) -> Optional[str]:
    pattern = r'isActive\:\s?(\w+)'
    is_active = get_regex_group(x=x, pattern=pattern, if_none=True)
    return is_active

def get_tmdl_relationship(tmdl: str) -> list[dict]:
    """
    Parse the relationships from a given TMDL content.

    Parameters
    ----------
    tmdl : str
        The content of a TMDL file.

    Returns
    -------
    list[dict]
        A list of dictionaries, each one representing a relationship.
        Each dictionary has the following keys:
            'name': str
                The name of the relationship.
            'fromTable': str
                The name of the table from which the relationship starts.
            'fromColumn': str
                The name of the column from which the relationship starts.
            'toTable': str
                The name of the table to which the relationship ends.
            'toColumn': str
                The name of the column to which the relationship ends.
            'crossFilteringBehavior': str
                The cross filtering behavior of the relationship.
                Can be 'oneDirection' or 'bothDirections'.
            'filteringSymbol': str
                The filtering symbol of the relationship.
                Can be '<' or '<>'.
            'fromCardinalitySymbol': str
                The cardinality symbol of the relationship.
                Can be '1' or '*'.
            'toCardinalitySymbol': str
                The cardinality symbol of the relationship.
                Can be '1' or '*'.
            'isActive': bool
                Whether the relationship is active or not.
    """

    splited = tmdl.split('relationship')[1:]
    relationships = []
    for x in splited:
        filtering_behavior = get_relationship_cross_filtering_behavior(x)
        from_cardinality = get_relationship_from_cardinality(x)
        to_cardinality = get_relationship_to_cardinality(x)

        filtering_cardinality = get_relationship_filter_cardinality(
            filtering_behavior,
            to_cardinality,
            from_cardinality
        )
        filtering_symbol, from_cardinality_symbol, to_cardinality_symbol = filtering_cardinality

        name = get_relationship_name(x)
        from_table, from_column = get_relationship_from_column(x)
        to_table, to_column = get_relationship_to_column(x)
        cross_filtering_behavior = get_relationship_cross_filtering_behavior(x)
        is_active = get_relationship_is_active(x)
        relationships.append({
            'name': name,
            'fromTable': from_table,
            'fromColumn': from_column,
            'toTable': to_table,
            'toColumn': to_column,
            'crossFilteringBehavior': cross_filtering_behavior,
            'filteringSymbol': filtering_symbol,
            'fromCardinalitySymbol': from_cardinality_symbol,
            'toCardinalitySymbol': to_cardinality_symbol,
            'isActive': is_active
        })
    return relationships

def get_expression_name(x: str) -> str:
    pattern = r"expression\s(.*)"
    name = str(get_regex_group(x=x, pattern=pattern)).split(' =')[0]
    return name

def get_expression_parameter_description(x: str, parameter: str) -> str:
    pattern = f'\/\/\/(.*)\nexpression\s({parameter})'
    search = re.search(pattern, x)
    if search is None:
        return ''
    return search.group(1).strip()

def get_expression_function_description(expression: str, pattern: str) -> str:
    return get_table_description(expression, pattern)

def is_expression_function(x: str):
    pattern = r'(annotation PBI_ResultType = Function)'
    return bool(get_regex_group(x=x, pattern=pattern))

def get_tmdl_expressions(tmdl: str) -> list[dict]:
    splited = [x for x in tmdl.split('\n') if x != '']
    no_tabs_idx = [idx 
        for (idx, value) in enumerate(splited)
        if value.count('\t') == 0
    ] + [len(splited)]
    sequences = list(zip(no_tabs_idx[0:-1], no_tabs_idx[1:]))
    expressions = []
    for begin, end in sequences:
        cur_list = splited[begin:end]
        cur = '\n'.join(splited[begin:end])
        expression_name = get_expression_name(cur)
        expression_lineage = get_lineage_tag(cur)
        if expression_name == 'None':
            continue
        is_function = is_expression_function(cur)
        if is_function:
            idx_next_level = [idx
                for (idx, value) in enumerate(cur_list[1:])
                if value.count('\t') == 1
            ][0] + 1
            raw_value = '\n'.join(cur_list[1:idx_next_level]).replace('\t\t', '')
            value = remove_doc_comments(raw_value).replace('```', '')
            description = get_expression_function_description(raw_value, r'@doc(.*?)(?=let)').replace('*/', '').strip()
        else:
            value = splited[begin].split('=')[1].split('meta')[0].strip()
            description = get_expression_parameter_description(x=tmdl, parameter=expression_name)
        expressions.append({
            'name': expression_name,
            'lineageTag': expression_lineage,
            'type': 'function' if is_function else 'expression',
            'expression': value,
            'description': description
        })
    return expressions

def build_table_path(semantic_model_path: str | Path) -> Path:
    new_path = Path(semantic_model_path)
    table_folder = Path.joinpath(new_path, 'definition', 'tables')
    return table_folder

def has_table_folder(semantic_model_path: str | Path) -> bool:
    new_path = Path(semantic_model_path)
    table_folder = build_table_path(new_path)
    if table_folder.exists():
        return True
    return False

def get_tmdl_table_files(semantic_model_path: str | Path) -> list[Path]:
    new_path = Path(semantic_model_path)
    table_folder = build_table_path(new_path)
    tables = list(table_folder.glob('*.tmdl'))
    return tables

def get_table_name(tmdl: str) -> Optional[str]:
    pattern = r"table\s+['\"]?([^'\"]+?)['\"]?\s*\n"
    name = get_regex_group(x=tmdl, pattern=pattern)
    return name

def get_table_content(tmdl: str) -> str:
    content = re.split(r'column|measure|partition|annotation|hierarchy', tmdl)[0]
    return content

def get_table_lineage_tag(tmdl: str) -> Optional[str]:
    table_content = get_table_content(tmdl)
    lineage_tag = get_lineage_tag(table_content)
    return lineage_tag

def check_table_property(tmdl: str, property: str) -> bool:
    table_content = get_table_content(tmdl)
    check = property in table_content
    return check

def get_annotations(x: str) -> Optional[list[dict]]:
    annotations = re.findall(r'annotation\s(\w+)\s?\=\s?(.*)', x)
    column_annotation = []
    if annotations == []:
        return None
    for x in annotations:
        column_annotation.append({'name': x[0], 'value': x[1]})
    return column_annotation

def get_table_columns_content(tmdl: str) -> list[str]:
    return get_table_property_content(tmdl, 'column')

def get_table_column_name_expression(x: str) -> tuple[Optional[str], Optional[str]]:
    pattern = r'(\w+)\s?\=?\s?(.*)$'
    x_splited = x.split('\n')[0]
    name = get_regex_group(x=x_splited, pattern=pattern, group=1)
    expression = get_regex_group(x=x_splited, pattern=pattern, group=2)
    return (name, expression)

def get_table_column_data_type(x: str) -> Optional[str]:
    pattern = r'dataType:\s?(\w+)'
    data_type = get_regex_group(x=x, pattern=pattern)
    return data_type

def get_table_summarize_by(x: str) -> Optional[str]:
    pattern = r'summarizeBy:\s?(\w+)'
    summarize_by = get_regex_group(x=x, pattern=pattern)
    return summarize_by

def get_table_source_column(x: str) -> Optional[str]:
    pattern = r'sourceColumn:\s?(\[?\w+\]?)'
    source_column = get_regex_group(x=x, pattern=pattern)
    return source_column

def get_table_column_type(x: str) -> Optional[str]:
    pattern = r'type:\s?(\w+)'
    type = get_regex_group(x=x, pattern=pattern)
    return type

def get_table_columns(tmdl: str) -> list[dict]:
    """
    Parse columns from the given tmdl string.

    Parameters
    ----------
    tmdl : str
        TMDL string

    Returns
    -------
    list[dict]
        A list of dictionaries containing column information. Each dictionary has the following keys:
        - name: str
        - expression: str
        - isHidden: bool
        - isNameInferred: bool
        - dataType: str | None
        - lineageTag: str | None
        - summarizeBy: str | None
        - sourceColumn: str | None
        - annotations: list[dict] | None
        - calculated: bool
    """
    content = get_table_columns_content(tmdl)
    columns = []
    for cur_value in content:
        name, expression = get_table_column_name_expression(cur_value)
        data_type = get_table_column_data_type(cur_value)
        is_hidden = 'isHidden' in cur_value
        is_name_inferred = 'isNameInferred' in cur_value
        lineage_tag = get_lineage_tag(cur_value)
        summarize_by = get_table_summarize_by(cur_value)
        source_column = get_table_source_column(cur_value)
        annotations = get_annotations(cur_value)
        columns.append({
            'name': name,
            'expression': expression,
            'isHidden': is_hidden,
            'isNameInferred': is_name_inferred,
            'dataType': data_type,
            'lineageTag': lineage_tag,
            'summarizeBy': summarize_by,
            'sourceColumn': source_column,
            'annotations': annotations,
            'calculated': str(source_column).startswith('[') or expression
        })
    return columns

def get_table_partition_content(tmdl: str) -> list[str]:
    return get_table_property_content(tmdl, 'partition')

def get_table_partition_name_type(x: str) -> tuple[Optional[str], Optional[str]]:
    pattern = r'(.*)\s?\=\s?(\w+)'
    x_splited = x.split('\n')[0]
    name = get_regex_group(x=x_splited, pattern=pattern, group=1)
    type = get_regex_group(x=x_splited, pattern=pattern, group=2)
    if name and type:
        return (name.strip(), type.strip())
    return (None, None)

def get_table_partition_mode(x: str) -> Optional[str]:
    pattern = r'mode:\s?(\w+)'
    mode = get_regex_group(x=x, pattern=pattern)
    return mode

def get_table_partition_source(x: str) -> Optional[str]:
    source_idx = re.search(r'source\s?=', x)
    if source_idx is None:
        return None
    else: 
        idx = source_idx.span()[1]
        source = x[idx:].replace('\t', '').replace('```', '').strip()
        return source

def get_table_partitions(tmdl: str) -> list[dict]:
    """
    Parse partitions from the given tmdl string.

    Parameters
    ----------
    tmdl : str
        TMDL string

    Returns
    -------
    list[dict]
        A list of dictionaries containing partition information. Each dictionary has the following keys:
        - name: str
        - type: str
        - mode: str
        - raw_expression: str
        - expression: str
        - description: str
    """
    content = get_table_partition_content(tmdl)
    partitions = []
    for cur_value in content:
        name, type = get_table_partition_name_type(cur_value)
        mode = get_table_partition_mode(cur_value)
        source = get_table_partition_source(cur_value)
        description = get_table_description(cur_value)
        source_without_docs = remove_doc_comments(source) if source else None
        partitions.append({
            'name': name,
            'type': type,
            'mode': mode,
            'raw_expression': source,
            'expression': source_without_docs,
            'description': description,
        })
    return partitions

def get_table_measure_content(tmdl: str) -> list[str]:
    return get_table_property_content(tmdl, 'measure')

def get_table_measure_name(x: str) -> str:
    name = x.split('=')[0].strip()
    return name

def get_table_measure_format_string(x: str) -> Optional[str]:
    pattern = r'formatString:\s?(.*)\n'
    format_string = get_regex_group(x=x, pattern=pattern)
    return format_string

def get_table_measure_display_folder(x: str) -> str:
    pattern = r'displayFolder:\s?(.*)\n'
    display_folder = get_regex_group(x=x, pattern=pattern, if_none='')
    return display_folder

def get_table_measure_expression(x: str) -> str:
    pattern = r'formatString|displayFolder|lineageTag|annotation'
    without_description = remove_doc_comments(x)
    raw_expression = re.split(pattern, without_description)[0].split('=')[1:]
    expression = '='.join(raw_expression).replace('\t', '').replace('```', '').strip()
    return expression

def get_table_measures(tmdl: str) -> list[dict]:
    """
    Parse measures from the given tmdl string.

    Parameters
    ----------
    tmdl : str
        TMDL string

    Returns
    -------
    list[dict]
        A list of dictionaries containing measure information. Each dictionary has the following keys:
        - name: str
        - formatString: str | None
        - annotations: list[str]
        - displayFolder: str | None
        - lineageTag: str | None
        - expression: str
        - description: str
        - references: list[str]
    """
    content = get_table_measure_content(tmdl)
    measures = []
    for cur_value in content:
        name = get_table_measure_name(cur_value)
        format_string = get_table_measure_format_string(cur_value)
        display_folder = get_table_measure_display_folder(cur_value)
        lineage_tag = get_lineage_tag(cur_value)
        annotations = get_annotations(cur_value)
        expression = get_table_measure_expression(cur_value)
        description = get_measure_description(cur_value)
        references = get_measure_references(expression)
        measures.append({
            'name': name,
            'formatString': format_string,
            'annotations': annotations,
            'displayFolder': display_folder,
            'lineageTag': lineage_tag,
            'expression': expression,
            'description': description,
            'references': references
        })
    return measures

def get_tmdl_table(tmdl: str) -> dict:
    """
    Reads a .tmdl file and returns a dictionary containing the table information

    Parameters
    ----------
    tmdl : str
        The contents of the .tmdl file

    Returns
    -------
    dict
        A dictionary containing the table information with the following keys:
            - name: The name of the table
            - lineageTag: The lineage tag of the table
            - isHidden: A boolean indicating if the table is hidden
            - isPrivate: A boolean indicating if the table is private
            - columns: A list of dictionaries containing the column information
            - measures: A list of dictionaries containing the measure information
            - partitions: A list of dictionaries containing the partition information
    """
    table_name = get_table_name(tmdl)
    lineage_tag = get_table_lineage_tag(tmdl)
    is_hidden = check_table_property(tmdl, 'isHidden')
    is_private = check_table_property(tmdl, 'isPrivate')
    table_columns = get_table_columns(tmdl)
    table_measures = get_table_measures(tmdl)
    table_partitions = get_table_partitions(tmdl)

    expression = table_partitions[0]['raw_expression']
    columns_with_descriptions = put_column_description(table_columns, expression)
    return {
        'name': table_name,
        'lineageTag': lineage_tag,
        'isHidden': is_hidden,
        'isPrivate': is_private,
        'columns': columns_with_descriptions,
        'measures': table_measures,
        'partitions': table_partitions
    }




from pathlib import Path
from textwrap import dedent

def is_pbip_file(path: str | Path) -> bool:
    new_path = Path(path)
    check = new_path.suffix == '.pbip'
    return check

def find_pbip_file(path: str | Path) -> list[Path]:
    new_path = Path(path)
    pbip_files = list(new_path.glob('*.pbip'))
    return pbip_files

def get_pbip_name(path: str | Path) -> str:
    new_path = Path(path)
    return '.'.join(new_path.name.split('.')[0:-1])

def get_component_path(path: str | Path, component: str) -> Path:
    new_path = Path(path)
    pbip_name = get_pbip_name(new_path)
    component_path = Path.joinpath(
        new_path.parent, 
        f'{pbip_name}.{component}'
    )
    return component_path

def get_semantic_model_path(path: str | Path) -> Path:
    return get_component_path(path, 'SemanticModel')

def get_report_path(path: str | Path) -> Path:
    return get_component_path(path, 'Report')

class PbipNotFoundError(Exception):
    def __init__(self, directory: str | Path):
        self.directory = directory
        message = dedent(f"""
            I'm trying to find a file with '.pbip' extension in the directory '{directory}', but I can't find one.
            Please, check if the directory is correct.
        """)
        super().__init__(message)

class MultiplePbipFilesError(Exception):
    def __init__(self, directory: str | Path, pbips: list[Path]):
        self.directory = directory
        message = dedent(f"""
            I find more than one '.pbip' file in the directory '{directory}'. But I expected only one. 
            The files I found are: {[x.name for x in pbips]}.
            Some alternatives could be:
                - Use only one .pbip file in the directory.
                - Pass .pbip path as an argument. 
                  For example: {Path.joinpath(Path(directory), pbips[0].name)}
        """)
        super().__init__(message)

class SemanticModelFolderNotFoundError(Exception):
    def __init__(self, path: Path):
        self.path = path
        message = dedent(f"""
            There isn't the folder {path}'.
        """)
        super().__init__(message)

class Pbip:
    """
    A class that represents pbip model.
    """
    def __init__(self, model: dict) -> None:
        """
        Initialize the Pbip object with a given model.

        Parameters
        ----------
        model : dict
            The model dictionary containing the data structure.
        """
        self._model = model

    @property
    def model(self) -> dict:
        return self._model
    
    def __getitem__(self, key):
        return self._model[key]

    @property
    def tables(self) -> list[dict]:
        return self.model['model']['tables']
    
    @property
    def expressions(self) -> list[dict]:
        return self.model['model']['expressions']
    
    @property
    def relationships(self) -> list[dict]:
        return self.model['model']['relationships']

    @relationships.setter
    def relationships(self, value: list[dict]):
        self.model['model']['relationships'] = value

    def get_table_field(self, table: str, field: str) -> list | None:
        """
        Get a field of a table.

        Parameters
        ----------
        table : str
            The name of the table.
        field : str
            The name of the field.

        Returns
        -------
        list | None
            The field of the table, or None if the table or field does not exist.
        """
        if table not in [x['name'] for x in self.tables]:
            return None
        for t in self.tables:
            if t['name'] == table and field in t:
                return t[field]
        return None

    def get_table_column(self, table: str, column: str) -> dict | None:
        """
        Get a column of a table.

        Parameters
        ----------
        table : str
            The name of the table.
        column : str
            The name of the column.

        Returns
        -------
        dict | None
            The column of the table, or None if the table or column does not exist.
        """
        if table not in [x['name'] for x in self.tables]:
            return None
        for t in self.tables:
            column_names = [x['name'] for x in t['columns']]
            if t['name'] == table and column in column_names:
                return t['columns'][column_names.index(column)]
        return None
    
    def get_table_measure(self, table: str, measure: str) -> dict | None:
        """
        Retrieve a specific measure from a table.

        Parameters
        ----------
        table : str
            The name of the table.
        measure : str
            The name of the measure.

        Returns
        -------
        dict | None
            The measure from the table, or None if the table or measure does not exist.
        """
        if table not in [x['name'] for x in self.tables]:
            return None
        for t in self.tables:
            measure_names = [x['name'] for x in t['measures']]
            if t['name'] == table and measure in measure_names:
                return t['measures'][measure_names.index(measure)]
        return None

    def get_fields(self, x: str) -> list[dict]:
        """
        Retrieve fields from tables based on the specified key.

        Parameters
        ----------
        x : str
            The key to access specific fields in the table, such as 'columns' or 'measures'.

        Returns
        -------
        list[dict]
            A list of dictionaries, each containing the fields from the tables along with the table name.
        """
        fields = []
        for table in self.tables:
            for column in table[x]:
                fields.append(column | {'table': table['name']})
        return fields

    @property
    def columns(self) -> list[dict]:
        return self.get_fields('columns')

    @property
    def measures(self) -> list[dict]:
        return self.get_fields('measures')




from collections import defaultdict
import re
from typing import Any


def get_regex_group(
    x: str, 
    pattern: str | re.Pattern[str],
    group: int = 1,
    if_none: Any = None
) -> Any:
    """
    Extracts a specified group from a regex match within a given string.

    Parameters
    ----------
    x : str
        The input string to search for the regex pattern.
    pattern : str or re.Pattern
        The regex pattern used to search within the input string.
    group : int, optional
        The group number to extract from the match, defaults to 1.
    if_none : Any, optional
        The value to return if no match is found, defaults to None.

    Returns
    -------
    Any
        The specified regex group if a match is found, otherwise returns `if_none`.
    """

    match = re.search(pattern, x)
    if match is None:
        return if_none
    return match.group(group)

def check_lines_max_length(x: str, max_length: int) -> list[tuple[str, str]]:
    """
    Checks if any line in a given string exceeds the maximum length.

    Parameters
    ----------
    x : str
        The input string to check.
    max_length : int
        The maximum length of a line.

    Returns
    -------
    list[tuple[str, str]]
        A list of tuples, where the first element of the tuple is the line
        number (1-indexed) and the second element is the line itself, if the
        line exceeds the maximum length. If no lines exceed the maximum length,
        an empty list is returned.
    """
    lines = x.split('\n')
    len_check = [
        (str(i + 1), x) 
        for i, x in enumerate(lines) 
        if len(x) > max_length
    ]
    return len_check

def quote_and_join(strings: list[str], sep=', '):
    """
    Joins a list of strings with a separator, single-quoting each string.

    Parameters
    ----------
    strings : list[str]
        The list of strings to join.
    sep : str, optional
        The separator to use when joining the strings, defaults to ', '.

    Returns
    -------
    str
        The joined string.
    """
    return sep.join(f"'{s}'" for s in strings)

def smart_join(strings: list[str]):
    """
    Joins a list of strings into a single string formatted in a human-readable way.

    Parameters
    ----------
    strings : list[str]
        The list of strings to join.

    Returns
    -------
    str
        A single string with the elements joined. If the list is empty, returns an 
        empty string. If there is one element, returns the element quoted. If there 
        are two elements, joins them with 'and'. For more than two elements, joins 
        all but the last with commas, and the last with 'and'.
    """
    if not strings:
        return ''
    elif len(strings) == 1:
        return f"'{strings[0]}'"
    elif len(strings) == 2:
        return quote_and_join(strings, ' and ')
    else:
        return quote_and_join(strings[:-1], ', ') + ' and ' + f"'{strings[-1]}'"

def sanitize_id(text: str) -> str:
    """
    Transforms a given string into a valid Markdown-friendly ID.
    
    - Converts to lowercase
    - Replaces spaces with hyphens
    - Removes special characters except hyphens and alphanumerics
    
    Parameters
    ----------
    text : str
        The string to sanitize
    
    Returns
    -------
    str
        The sanitized string
    """
    text = text.lower() 
    text = re.sub(r'\s+', '-', text)
    text = re.sub(r'[^a-z0-9\-]', '', text)
    return text

def fix_duplicate_ids(text: str) -> str:
    """
    Finds duplicate HTML id attributes and markdown links (e.g. [text](#id))
    and fixes them so that each occurrence is unique.

    For HTML tag attributes (like id="Duplicate"), the first occurrence remains
    unchanged, the second becomes "Duplicate1", the third "Duplicate2", etc.

    For markdown links, we assume that the intended target should follow the
    same order. For example, the first markdown link referencing "#Duplicate" will
    be changed to point to the first occurrence (i.e. "#Duplicate"), the second will
    be changed to "#Duplicate1", etc.

    Code blocks (demarcated by "```") are left unmodified.

    Parameters
    ----------
    text : str
        The rendered text to fix duplicate ids and markdown links in.

    Returns
    -------
    str
        The rendered text with duplicate ids and markdown links fixed.
    """
    id_mapping = defaultdict(list)
    counts: dict = defaultdict(int)
    link_counts: dict = defaultdict(int)

    id_attr_pattern = re.compile(r'(\bid=")([^"]+)(")')
    md_link_pattern = re.compile(r'(\]\(#)([^)]+)(\))')

    def replace_html_id(match):
        prefix, orig_id, suffix = match.groups()
        cnt = counts[orig_id]
        new_id = orig_id if cnt == 0 else f"{orig_id}{cnt}"
        counts[orig_id] += 1
        id_mapping[orig_id].append(new_id)
        return f'{prefix}{new_id}{suffix}'

    def replace_md_link(match):
        prefix, orig_id, suffix = match.groups()
        if orig_id not in id_mapping or not id_mapping[orig_id]:
            return match.group(0)
        cnt = link_counts[orig_id]
        if cnt < len(id_mapping[orig_id]):
            new_target = id_mapping[orig_id][cnt]
        else:
            new_target = id_mapping[orig_id][0]
        link_counts[orig_id] += 1
        return f'{prefix}{new_target}{suffix}'

    code_block_pattern = re.compile(r'(```[\s\S]*?```)', re.MULTILINE)
    segments = code_block_pattern.split(text)

    for i, segment in enumerate(segments):
        if i % 2 == 0:
            segment = id_attr_pattern.sub(replace_html_id, segment)
            segment = md_link_pattern.sub(replace_md_link, segment)
            segments[i] = segment

    fixed_text = "".join(segments)
    return fixed_text




from __future__ import annotations
import textwrap
import json
from pathlib import Path
from pbipinspect.expectations import ExpectMessage
from pbipinspect.pbip import (
    is_pbip_file, 
    find_pbip_file,
    get_pbip_name, 
    get_semantic_model_path,
    get_report_path,
    MultiplePbipFilesError,
    Pbip,
    PbipNotFoundError,
    SemanticModelFolderNotFoundError
)
from pbipinspect.parse.tmdl import (
    get_tmdl_expressions,
    get_tmdl_expression_file,
    get_tmdl_relationship,
    get_tmdl_relationship_file,
    get_tmdl_table,
    get_tmdl_table_files,
    has_table_folder
)
from pbipinspect.parse.tmsl import (
    build_tmsl_path,
    clean_tmsl,
    is_TMSL
)
from pbipinspect.report.templates import DEFAULT_TEMPLATE
from pbipinspect.utils import (
    fix_duplicate_ids,
    sanitize_id
)
from typing import Any, Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from jinja2 import Template

class PbipInspect:
    """
    A class for inspecting .pbip files and their associated TMDL and TMSL models.
    """
    def __init__(
        self, 
        name: str | Path,
        semantic_folder: str | Path,
        report_folder: str | Path
    ) -> None:
        """
        Initialize the PbipInspect object with a given name, semantic folder and report folder.

        Parameters
        ----------
        name : str | Path
            The name of the .pbip file.
        semantic_folder : str | Path
            The path of the semantic model folder.
        report_folder : str | Path
            The path of the report folder.

        Returns
        -------
        None
        """
        self.name = name
        self.semantic_folder = semantic_folder
        self.report_folder = report_folder
        self._pbip = self.parse_pbip()
        self.steps: list[Callable] = []
        self.expects: list[ExpectMessage] = []
        self.metadata: dict = {}

    def __repr__(self) -> str:
        return f"PbipInspect(name='{self.name}', semantic_folder='{self.semantic_folder}', report_folder='{self.report_folder}')"

    @property
    def pbip(self) -> Pbip:
        return self._pbip

    def parse_tmdl_tables(self) -> list[dict] | None:
        """
        Parse all .tmdl files in the 'definition/tables' folder.

        Returns
        -------
        list[dict] | None
            If the 'definition/tables' folder does not exists, return None.
            Otherwise, return a list of dictionaries, each one representing a table.
        """
        folder = self.semantic_folder
        check = has_table_folder(folder)
        if not check:
            return None
        tables = get_tmdl_table_files(folder)
        parsed_tables = []
        for table in tables:
            content = table.read_text(encoding='utf-8')
            tmdl_parsed = get_tmdl_table(content)
            parsed_tables.append(tmdl_parsed)
        return parsed_tables

    def parse_tmdl_relationships(self) -> list[dict]:
        """
        Parse the relationships from the 'relationships.tmdl' file.

        Returns
        -------
        list[dict] | None
            If the 'definition/tables' folder does not exist, return None.
            Otherwise, return a list of dictionaries with relationship details.
        """
        folder = self.semantic_folder
        check = has_table_folder(folder)
        if not check:
            return []
        relationship = get_tmdl_relationship_file(folder)
        if Path(relationship).is_file() == False: 
            return []
        content = relationship.read_text(encoding='utf-8')
        parsed_relationship = get_tmdl_relationship(content)
        return parsed_relationship
    
    def parse_tmdl_expressions(self) -> list[dict]:
        """
        Parse the expressions from the 'expressions.tmdl' file.

        Returns
        -------
        list[dict] | None
            If the 'definition/tables' folder does not exist, return None.
            Otherwise, return a list of dictionaries with expression details.
        """
        folder = self.semantic_folder
        check = has_table_folder(folder)
        if not check:
            return []
        expression = get_tmdl_expression_file(folder)
        if Path(expression).is_file() == False: 
            return []
        content = expression.read_text(encoding='utf-8')
        parsed_expression = get_tmdl_expressions(content)
        return parsed_expression

    #TODO: need to create a real tmdl parser
    def parse_tmdl(self) -> dict:
        """
        Parse the TMDL model from the semantic folder.

        Returns
        -------
        dict
            A dictionary representing the model, containing lists of tables and relationships.
        """
        final_model: dict = {
            'model': {
                'tables': [],
                'relationships': [],
                'expressions': []
            }
        }
        final_model['model']['tables'] = self.parse_tmdl_tables()
        final_model['model']['relationships'] = self.parse_tmdl_relationships()
        final_model['model']['expressions'] = self.parse_tmdl_expressions()
        return final_model

    def parse_tmsl(self) -> dict:
        """
        Parse the TMSL model from the semantic folder.

        Returns
        -------
        dict
            A dictionary representing the model, containing lists of tables and relationships.
        """
        model_bim = build_tmsl_path(self.semantic_folder)
        with open(model_bim, 'r', encoding='utf-8') as f:
            tmsl = json.load(f)
        final_model = clean_tmsl(tmsl)
        return final_model

    def parse_pbip(self) -> Pbip:
        """
        Parse the PBIP model from the semantic folder.

        Returns
        -------
        Pbip
            An object representing the parsed model, containing the model and a method to run expectations.
        """
        tmsl_check = is_TMSL(self.semantic_folder)
        parse = self.parse_tmsl if tmsl_check else self.parse_tmdl
        final_model = parse()
        return Pbip(final_model)

    def expectations(self, steps: list[Callable]):
        """
        Register a list of expectation functions to run against the model.

        Parameters
        ----------
        steps : list[Callable]
            A list of functions that take a Pbip object as an argument and return a list of ExpectMessage.
        """
        self.steps = steps
    
    def run_expectations(self):
        """
        Executes all registered expectation functions against the model.

        The method iterates over the list of expectation functions stored in `self.steps` 
        and applies each function to the model. It collects the results, which are lists 
        of `ExpectMessage` dictionaries, and flattens them into a single list stored in 
        `self.expects`.

        Returns
        -------
        None
        """
        results = []
        for expect in self.steps:
            results.append(expect(self.pbip))
        self.expects = [x for y in results for x in y]
    
    def add_metadata(self, metadata: dict):
        """
        Add metadata to the report.

        Parameters
        ----------
        metadata : dict
            A dictionary of metadata to be added to the report.
        """
        self.metadata = metadata
    
    @staticmethod
    def parse_relationships_to_mermaid(content: dict) -> str:
        """
        Given a JSON content describing Power BI relationships,
        this function returns a string in Mermaid flowchart format that shows each table (as a subgraph)
        with its columns, and draws arrows representing the relationships with cardinality labels.

        Parameters
        ----------
        content : dict
            A dictionary representing the model, containing lists of tables and relationships.

        Returns
        -------
        str
            A string in Mermaid flowchart format.
        """
        relationships = content.get('relationships', [])
        table_columns: dict = {}
        for rel in relationships:
            from_table, from_column = rel.get('fromTable'), rel.get('fromColumn')
            to_table, to_column = rel.get('toTable'), rel.get('toColumn')
            if from_table and from_column:
                table_columns.setdefault(from_table, set()).add(from_column)
            if to_table and to_column:
                table_columns.setdefault(to_table, set()).add(to_column)
        
        table_subgraph_ids = {}
        node_ids = {}
        subgraph_lines = []
        node_counter, subgraph_counter = 1, 1

        for table in sorted(table_columns.keys()):
            columns = table_columns[table]
            subgraph_id = f"s{subgraph_counter}"
            table_subgraph_ids[table] = subgraph_id
            subgraph_counter += 1
            
            sub_lines = [f'subgraph {subgraph_id}["{table}"]']
            for col in sorted(columns):
                node_id = f"n{node_counter}"
                node_ids[(table, col)] = node_id
                node_counter += 1
                sub_lines.append(f'{node_id}["{col}"]')
            sub_lines.append("end")
            subgraph_lines.append("\n".join(sub_lines))
        
        relationship_lines = []
        for rel in relationships:
            from_table, from_column = rel.get('fromTable'), rel.get('fromColumn')
            to_table, to_column = rel.get('toTable'), rel.get('toColumn')
            from_node = node_ids.get((from_table, from_column))
            to_node = node_ids.get((to_table, to_column))
            from_card, to_card = rel.get('fromCardinalitySymbol', ''), rel.get('toCardinalitySymbol', '')
            from_card_mermaid = f'\\{from_card}' if from_card == '*' else from_card
            filtering_symbol = rel.get('filteringSymbol')
            line = f'{from_node} -- {from_card_mermaid} {filtering_symbol} {to_card} --- {to_node}'
            relationship_lines.append(line)
        
        mermaid_lines = ['flowchart LR']
        mermaid_lines.extend(subgraph_lines)
        mermaid_lines.extend(relationship_lines)
        return "\n".join(mermaid_lines)
    
    def build_report(
        self, 
        report_template: str = DEFAULT_TEMPLATE,
        render: Callable[[Template, tuple[Any], dict[str, Any]], str] | None = None,
        *args: Any,
        **kwargs: Any
    ):
        from jinja2 import Template
        if self.steps and not self.expects:
            self.run_expectations()

        template = Template(source=textwrap.dedent(report_template))
        if not render:
            report = template.render(
                metadata=self.metadata,
                tables=self.pbip.tables,
                mermaid_relationships=self.parse_relationships_to_mermaid(self.pbip.model['model']),
                expressions = self.pbip.expressions,
                expects=self.expects,
                sanitize_id=sanitize_id
            )
        else:
            report = render(template, *args, **kwargs)
        return fix_duplicate_ids(report)


def create_inspect(path: str | Path) -> PbipInspect:
    """
    Create a PbipInspect object from a given path, which can be a .pbip file or a directory 
    containing .pbip files. Validates the presence and uniqueness of .pbip files, and verifies 
    the existence of required folders.

    Parameters
    ----------
    path : str | Path
        The path to a specific .pbip file or a directory that may contain .pbip files.

    Returns
    -------
    PbipInspect
        An instance of PbipInspect initialized with the .pbip file's name, semantic folder, 
        and report folder.

    Raises
    ------
    PbipNotFoundError
        If no .pbip file is found in the specified directory.
    MultiplePbipFilesError
        If more than one .pbip file is found in the specified directory.
    SemanticModelFolderNotFoundError
        If the semantic folder does not exist.
    """
    check_pbip = is_pbip_file(path)
    pbips = [Path(path)] if check_pbip else find_pbip_file(path)
    
    if not pbips:
        raise PbipNotFoundError(path)
    
    if len(pbips) > 1:
        raise MultiplePbipFilesError(path, pbips)
    
    pbip = pbips[0]
    pbip_name = get_pbip_name(pbip)
    semantic_folder = get_semantic_model_path(pbip)

    if not semantic_folder.is_dir():
        raise SemanticModelFolderNotFoundError(semantic_folder)

    report_folder = get_report_path(pbip)
    inspect = PbipInspect(pbip_name, semantic_folder, report_folder)
    return inspect




from pathlib import Path
from pbipinspect.parse.utils import (
    get_measure_references,
    put_column_description,
    get_measure_description,
    get_relationship_filter_cardinality,
    get_table_description,
    remove_doc_comments
)

def build_tmsl_path(semantic_model_path: str | Path) -> Path:
    new_path = Path(semantic_model_path)
    model = Path.joinpath(new_path, 'model.bim')
    return model

def is_TMSL(semantic_model_path: str | Path) -> bool:
    model_bim = build_tmsl_path(semantic_model_path)
    return model_bim.exists()

def clean_tmsl_columns(columns: list[dict]) -> list[dict]:
    """
    Clean columns from the given TMSL model.

    Parameters
    ----------
    columns : list[dict]
        The list of columns to clean

    Returns
    -------
    list[dict]
        A list of cleaned column dictionaries. The dictionaries have the following keys:
        - name: str
        - expression: str | None
        - isHidden: bool
        - isNameInferred: bool
        - dataType: str
        - lineageTag: str
        - summarizeBy: str
        - sourceColumn: str | None
        - annotations: list[str] | None
        - calculated: bool
        - description: str
    """
    cleaned: list[dict] = []
    if not columns:
        return cleaned
    for column in columns:
        cleaned.append({
            'name': column['name'],
            'expression': column.get('expression', ''),
            'isHidden': column.get('isHidden', False),
            'isNameInferred': column.get('isHidden', False),
            'dataType': column['dataType'],
            'lineageTag': column['lineageTag'],
            'summarizeBy': column['summarizeBy'],
            'sourceColumn': column.get('sourceColumn'),
            'annotations': column.get('annotations'),
            'calculated': str(column.get('sourceColumn')).startswith('[') or bool(column.get('expression')),
            'description': column.get('description', '')
        })
    return cleaned

def clean_tmsl_partitions(partitions: list[dict]) -> list[dict]:
    """
    Clean the given list of partition dictionaries.

    Parameters
    ----------
    partitions : list[dict]
        The list of partition dictionaries to clean

    Returns
    -------
    list[dict]
        A list of cleaned partition dictionaries. The dictionaries have the following keys:
        - name: str
        - type: str
        - mode: str
        - raw_expression: str
        - expression: str
        - description: str
    """
    cleaned = []
    for partition in partitions:
        expression = partition['source']['expression']
        if isinstance(expression, str):
            expression = [expression]
        expression = '\n'.join(expression).strip()
        cleaned.append({
            'name': partition['name'],
            'type': partition['source']['type'],
            'mode': partition['mode'],
            'raw_expression': expression,
            'expression': remove_doc_comments(expression),
            'description': get_table_description(expression)
        })
    return cleaned

def clean_tmsl_measures(measures: list[dict]) -> list[dict]:
    """
    Clean measures from the given TMSL model.

    Parameters
    ----------
    measures : list[dict]
        The list of measures to clean.

    Returns
    -------
    list[dict]
        A list of cleaned measure dictionaries. Each dictionary contains the following keys:
        - name: str
        - lineageTag: str
        - annotations: list[str] | None
        - displayFolder: str
        - expression: str
        - formatString: str | None
        - description: str
        - references: list[str]
    """
    cleaned: list[dict] = []
    if not measures:
        return cleaned
    for measure in measures:
        expression = measure['expression']
        if isinstance(expression, str):
            expression = [expression]
        expression = '\n'.join(expression).strip()
        cleaned.append({
            'name': measure['name'],
            'lineageTag': measure['lineageTag'],
            'annotations': measure.get('annotations'),
            'displayFolder': measure.get('displayFolder', ''),
            'expression': remove_doc_comments(expression),
            'formatString': measure.get('formatString'),
            'description': get_measure_description(expression),
            'references': get_measure_references(expression)
        })
    return cleaned

def clean_tmsl_tables(tables: list[dict]) -> list[dict]:
    """
    Clean tables from the given TMSL model.

    Parameters
    ----------
    tables : list[dict]
        The list of tables to clean.

    Returns
    -------
    list[dict]
        A list of cleaned table dictionaries. Each dictionary contains the following keys:
        - name: str
        - lineageTag: str
        - isHidden: bool
        - isPrivate: bool
        - columns: list[dict]
        - measures: list[dict]
        - partitions: list[dict]
    """
    cleaned = []
    for table in tables:
        expression = table['partitions'][0]['source']['expression']
        if isinstance(expression, str):
            expression = [expression]
        expression = '\n'.join(expression).strip()
        columns = table.get('columns', [])
        columns_with_descriptions = put_column_description(columns, expression)
        
        cleaned.append({
            'name': table['name'],
            'lineageTag': table['lineageTag'],
            'isHidden': table.get('isHidden', False),
            'isPrivate': table.get('isPrivate', False),
            'columns': clean_tmsl_columns(columns_with_descriptions),
            'measures': clean_tmsl_measures(table.get('measures', [])),
            'partitions': clean_tmsl_partitions(table['partitions'])
        })
    return cleaned

def clean_tmsl_relationships(relationships: list[dict]) -> list[dict]:
    """
    Clean relationships from the given TMSL model.

    Parameters
    ----------
    relationships : list[dict]
        The list of relationships to clean.

    Returns
    -------
    list[dict]
        A list of cleaned relationship dictionaries. Each dictionary contains the following keys:
        - name: str
        - fromTable: str
        - fromColumn: str
        - toTable: str
        - toColumn: str
        - crossFilteringBehavior: str
        - filteringSymbol: str
        - fromCardinalitySymbol: str
        - toCardinalitySymbol: str
        - isActive: bool
    """
    cleaned = []
    for relationship in relationships:
        filtering_behavior = relationship.get('crossFilteringBehavior', 'singleDirection')
        to_cardinality = relationship.get('toCardinality')
        from_cardinality = relationship.get('fromCardinality')

        filtering_cardinality = get_relationship_filter_cardinality(
            filtering_behavior,
            to_cardinality,
            from_cardinality
        )
        filtering_symbol, from_cardinality_symbol, to_cardinality_symbol = filtering_cardinality

        cleaned.append({
            'name': relationship['name'],
            'fromTable': relationship['fromTable'],
            'fromColumn': relationship['fromColumn'],
            'toTable': relationship['toTable'],
            'toColumn': relationship['toColumn'],
            'crossFilteringBehavior': filtering_behavior,
            'filteringSymbol': filtering_symbol,
            'fromCardinalitySymbol': from_cardinality_symbol,
            'toCardinalitySymbol': to_cardinality_symbol,
            'isActive': relationship.get('isActive', True)
        })
    return cleaned

def clean_tmsl_expressions(expressions: list[dict]) -> list[dict]:
    """
    Clean expressions from the given TMSL model.

    Parameters
    ----------
    expressions : list[dict]
        The list of expressions to clean.

    Returns
    -------
    list[dict]
        A list of cleaned expression dictionaries. Each dictionary contains the following keys:
        - name: str
        - expression: str
        - description: str
    """
    cleaned = []
    for expression in expressions:
        is_function = any([True for x in expression['annotations'] if x['value'] == 'Function'])
        if not is_function:
            expr = expression['expression'].split('meta')[0].strip()
        else:
            expr = '\n'.join(expression['expression']).strip()
        cleaned.append({
            'name': expression['name'],
            'lineageTag': expression['lineageTag'],
            'expression': expr,
            'description': expression.get('description'),
            'type': 'function' if is_function else 'parameter'
        })
    return cleaned

def clean_tmsl(tmsl_model: dict) -> dict:
    """
    Clean the given TMSL model.

    Parameters
    ----------
    tmsl_model : dict
        The TMSL model to clean.

    Returns
    -------
    dict
        A cleaned version of the given TMSL model.
    """
    model = tmsl_model['model']
    tables = model.get('tables', [])
    relationships = model.get('relationships', [])
    expressions = model.get('expressions', [])
    cleaned_tables = clean_tmsl_tables(tables)
    cleaned_relationships = clean_tmsl_relationships(relationships)
    cleaned_expressions = clean_tmsl_expressions(expressions)
    new_model = {
        'model': {
            'tables': cleaned_tables,
            'relationships': cleaned_relationships,
            'expressions': cleaned_expressions
        }
    }
    return new_model
